@@ -9,7 +9,7 @@ except ImportError:
     # fallback for Self in python 3.10
     from typing import TypeVar
     Self = TypeVar("Self", bound="Cube")
-from typing import Tuple, Union
+from typing import Tuple, Union, Callable
 import pandas as pd
 import numbers
 import sys
@@ -239,7 +239,12 @@ class Cube:
             None
         """
         def make_iri(row):
-            return self._cube_uri + "/observation/" + "_".join([quote(str(row[key_dim])) for key_dim in self._key_dimensions])
+            parts = [
+            quote(str(row[key_dim]), safe="")  # safe="" means *everything* that’s not unreserved will be encoded
+            for key_dim in self._key_dimensions
+            ]
+            return f"{self._cube_uri}/observation/{'_'.join(parts)}"
+        
         self._dataframe['obs-uri'] = self._dataframe.apply(
             make_iri, axis=1
         )
@@ -294,11 +299,25 @@ class Cube:
 
                         # Perform the {} placeholder replacement with the column values, for each row
                         self._dataframe[dim_name] = self._dataframe.apply(lambda row: self._replace_placeholders(row, repl), axis=1)
+                    case "function":
+                        func = self._load_function_via_exec(mapping.get("filepath"), mapping.get("function-name"))
+                        self._dataframe[dim_name] = self._dataframe[dim_name].map(func)
                         
                 value_type = mapping.get("value-type", 'Shared')
                 assert value_type in ['Shared', 'Literal']
                 self._dataframe[dim_name] = self._dataframe[dim_name].map(lambda v: URIRef(v) if value_type == "Shared" else Literal(v))
 
+    @staticmethod
+    def _load_function_via_exec(filepath: str, function_name: str) -> Callable:
+        namespace = {}
+        with open(filepath, "r") as f:
+            code = f.read()
+        exec(code, namespace)  # Execute all code in the file in this namespace
+        func = namespace.get(function_name)
+        if func is None:
+            raise ValueError(f"Function '{function_name}' not found in {filepath}")
+        return func
+    
     def _write_dcat_contact_point(self, contact_dict: dict) -> BNode | URIRef:
         """Writes a contact point to the graph.
         
@@ -631,15 +650,18 @@ class Cube:
         
         self._graph.add((dim_node, SH.path, URIRef(self._base_uri + dim_dict.get("path"))))
 
+        if dim_dict.get("datatype") == "URI":
+            self._graph.add((dim_node, SH.nodeKind, SH.IRI))
+        else:
+            self._graph.add((dim_node, SH.nodeKind, SH.Literal))
+            self._graph.add((dim_node, SH.datatype, XSD[dim_dict.get("datatype")]))
+
         match dim_dict.get("dimension-type"):
             case "Key Dimension":
                 self._graph.add((dim_node, RDF.type, CUBE.KeyDimension))
-                self._graph.add((dim_node, SH.nodeKind, SH.IRI))
-                self._graph.add((dim_node, SH.nodeKind, SH.IRI))
                 
             case "Measure Dimension":
                 self._graph.add((dim_node, RDF.type, CUBE.MeasureDimension))
-                self._graph.add((dim_node, SH.nodeKind, SH.Literal))
             
             case "Annotation":
                 self._graph.add((dim_node, SH.nodeKind, SH.Literal))
@@ -678,17 +700,17 @@ class Cube:
             case "nominal":
                 self._graph.add((dim_node, QUDT.scaleType, QUDT.NominalScale))
                 if dim_dict.get("dimension-type") == "Key Dimension":
-                    self._add_sh_list(dim_node, values)
+                    self._add_sh_list(dim_dict, dim_node, values)
             case "ordinal":
                 self._graph.add((dim_node, QUDT.scaleType, QUDT.OrdinalScale))
                 if dim_dict.get("dimension-type") == "Key Dimension":
-                    self._add_sh_list(dim_node, values)
+                    self._add_sh_list(dim_dict, dim_node, values)
             case "interval":
                 self._graph.add((dim_node, QUDT.scaleType, QUDT.IntervalScale))
-                self._add_min_max(dim_node, values)
+                self._add_min_max(dim_dict, dim_node, values)
             case "ratio":
                 self._graph.add((dim_node, QUDT.scaleType, QUDT.RatioScale))
-                self._add_min_max(dim_node, values)
+                self._add_min_max(dim_dict, dim_node, values)
             case _ as unrecognized:
                 print(f"Scale Type '{unrecognized}' is not recognized")
         
@@ -705,7 +727,7 @@ class Cube:
                     case "temporal":
                         data_kind_node = BNode()
                         self._graph.add((data_kind_node, RDF.type, TIME.GeneralDateTimeDescription))
-                        self._graph.add((data_kind_node, TIME.unitType, TIME.unitYear))
+                        self._graph.add((data_kind_node, TIME.unitType, TIME["unit" + data_kind.get("unit").capitalize()]))
                         self._graph.add((dim_node, META.dataKind, data_kind_node))
                     case "spatial-shape":
                         data_kind_node = BNode()
@@ -856,7 +878,7 @@ class Cube:
 
         return context_node
     
-    def _add_sh_list(self, dim_node: BNode, values: pd.Series):
+    def _add_sh_list(self, dimension_dict: dict, dim_node: BNode, values: pd.Series):
         """Add a SHACL list of all unique values to the given dimension node.
         
             Args:
@@ -868,10 +890,13 @@ class Cube:
         """
         list_node = BNode()
         unique_values = values.unique()
-        Collection(self._graph, list_node, [URIRef(vl) for vl in unique_values])
+        if dimension_dict.get("datatype") == "URI":
+            Collection(self._graph, list_node, [URIRef(vl) for vl in unique_values])
+        else:
+            Collection(self._graph, list_node, [Literal(vl, datatype=XSD[dimension_dict.get("datatype")]) for vl in unique_values])
         self._graph.add((dim_node, URIRef(SH + "in"), list_node))
 
-    def _add_min_max(self, dim_node: BNode, values: pd.Series):
+    def _add_min_max(self, dim_dict: dict, dim_node: BNode, values: pd.Series):
         """Add minimum and maximum values to the given dimension node.
         
             Args:
@@ -884,8 +909,10 @@ class Cube:
         # todo: case of cube.Undefined should be covered
         _min = values.min()
         _max = values.max()
-        self._graph.add((dim_node, SH.min, Literal(_min)))
-        self._graph.add((dim_node, SH.max, Literal(_max)))
+
+        _datatype = dim_dict.get("datatype")
+        self._graph.add((dim_node, SH.minInclusive, Literal(_min, datatype=XSD[_datatype])))
+        self._graph.add((dim_node, SH.maxInclusive, Literal(_max, datatype=XSD[_datatype])))
 
     @staticmethod
     def _sanitize_value(value, datatype, lang=None) -> Literal|URIRef:
